function engine = jtree_C_inf_engine(bnet, varargin)
% JTREE_C_INF_ENGINE Junction tree inference engine, C implemenation
% engine = jtree_C_inf_engine(bnet, ...)
%
% This is very similar to jtree_inf_engine, except all hidden nodes must be discrete.
% Also, one can currently only compute marginals on single nodes and families,
% not on arbitrary subsets of nodes.
% The C code was implemented by Wei Hu.

% The difference between jtree and jtree_C is : jtree_C use the C mex to do the
% enter_soft_evidence, and it's output clpot is cell of array, not cell of object.


% set default params
N = length(bnet.dag);
onodes = [];
clusters = {};
root = N;
stages = { 1:N };

if nargin >= 2
  args = varargin;
  nargs = length(args);
  if ~isstr(args{1})
    if nargs >= 1 & ~isempty(args{1}), onodes = args{1}; end
    if nargs >= 2 & ~isempty(args{2}), clusters = args{2}; end
    if nargs >= 3 & ~isempty(args{3}), stages = args{3}; end
    if nargs >= 4 & ~isempty(args{4}), root = args{4}; end
  else
    for i=1:2:nargs
      switch args{i},
       case 'observed', onodes = args{i+1}; 
       case 'clusters', clusters = args{i+1}; 
       case 'root',     root = args{i+1}; 
       case 'stages',   stages = args{i+1}; 
       otherwise,  
	error(['invalid argument name ' args{i}]);       
      end
    end
  end
end

engine = init_fields;
engine = class(engine, 'jtree_C_inf_engine', inf_engine(bnet));
engine.observed = onodes;

[engine.jtree, dummy, engine.cliques, B, w] = dag_to_jtree(bnet, onodes, stages, clusters);
engine.cliques_bitv = B;
engine.clique_weight = w;
C = length(engine.cliques);
engine.clpot = cell(1,C);

% Compute the separators between connected cliques.
[is,js] = find(engine.jtree > 0);
engine.separator = cell(C,C);
for k=1:length(is)
  i = is(k); j = js(k);
  engine.separator{i,j} = find(B(i,:) & B(j,:)); % intersect(cliques{i}, cliques{j});
end

% A node can be a member of many cliques, but is assigned to exactly one, to avoid
% double-counting its CPD. We assign node i to clique c if c is the "lightest" clique that
% contains i's family, so it can accomodate its CPD.

engine.clq_ass_to_node = zeros(1, N);
for i=1:N
  %c = clq_containing_nodes(engine, family(bnet.dag, i));
  clqs_containing_family = find(all(B(:,family(bnet.dag, i)), 2)); % all selected columns must be 1
  c = clqs_containing_family(argmin(w(clqs_containing_family)));  
  engine.clq_ass_to_node(i) = c; 
end

% Make the jtree rooted, so there is a fixed message passing order.
engine.root_clq = clq_containing_nodes(engine, root);
if engine.root_clq <= 0
  error(['no clique contains ' num2str(root)]);
end

[engine.jtree, engine.preorder, engine.postorder] = mk_rooted_tree(engine.jtree, engine.root_clq);

% collect 
engine.postorder_parents = cell(1,length(engine.postorder));
for n=engine.postorder(:)'
  engine.postorder_parents{n} = parents(engine.jtree, n);
end
% distribute
engine.preorder_children = cell(1,length(engine.preorder));
for n=engine.preorder(:)'
  engine.preorder_children{n} = children(engine.jtree, n);
end

  

%%%%%%%%

function engine = init_fields()

engine.jtree = [];
engine.cliques = [];
engine.separator = [];
engine.cliques_bitv = [];
engine.clique_weight = [];
engine.clpot = [];
engine.clq_ass_to_node = [];
engine.root_clq = [];
engine.preorder = [];
engine.postorder = [];
engine.preorder_children = [];
engine.postorder_parents = [];
engine.maximize = [];
engine.evidence = [];
engine.observed = [];
