
function [outputMats ,Hyp,NeighborPurity,KnnError,DevelopementStats]= DistBoost(data,chunklets,anti,labels,BSParams)

% DISTBOOST - trains a distance function using datapoints and equivalence constraints.
%
%    Description:
% 	 [FinalHyp,NZ,NP,KNN,DevelopementStats,MogParams]= DistBoost(data,labels,chunklets,anti,BSParams)
%	
%
%       the parameters are:
%
% 	  data =       The training data matrix data, each datapoint is a row vector.
%
%  	  labels =     The training labels (optional). If passed, used for various evaluation
%                        statistics. default set to [].
%
%
%         chunklets =  A vector denoting positive equivalence constraints in the followin format: 
%                        (1) The vector is 1*N where N is the data size.
%                        (2) The value at index i denotes the chunklet to which point i belongs to
%                            this value can be -1 (unconstrained point) or 1...[#of chunklets]. see
%                            example file for a detailed example.
%
% 	  anti =       A matrix of S*2 where S is the number of negative equivalence constraints. Each
%                      row contains the indices of two negatively constrained datapoints.
%
% 	The BSParams struct (optional) must contain all the relevant parameters to tune DistBoost. 
%
%  	The fields in BSParams are
%	
% 
% 	  WeakParams = The weak learner parmeters - see initWeakParams for details. in this
% 	               struct you would usually like to  set following fields:
%                            M =  (optional) the number of models used by the weak learner GMM 
%                                 (should be >= true # of classes). 
%                                 defaults settings: If labels are passed will use true # of
%                                 classes. If not, will simply set to 5 and a warning message
%                                 will be printed.
%
%
%                            num_itertations = number of EM iterations in each weak learner ...
%				 (optional). Default value is set to 6.
%
%                            number_of_starts = The number of different initial conditions
%                                               sampled from which the weak learner chooses its initial parameters
%                                               (optional). Default is set to 3 (to allow bold hyps to be selected).	    
%
%
% 	  T =          The number of boosting iterations (optional). Default value set to 15.
%	  
% 	  numberOfCalls =  The number of attempts to recall the weak learner in case it
%                             fails to produce a valid hypothesis (optional). Default value set to 8.
%
% 	  simple_alpha =   If set to zero will compute step size using line-search. Default value
% 	                   set to 1 (like standard AdaBoost update).
%
% 	  Lambda =         Set the weight of the unsupervised vs. supervised points.  Default value set to 1.
%         
%         InitialPairW =   Allows to set initial weights on constraints before beginning the
%                          boosting process. if not set, the weights are initialized uniformly on all pairs.
%	  
%
%         printStatsFlag   =   If 1 will print boosting developmet stats to the screen, if zero no printing.
%
%         evaluateHypFlag  =   If 1 will compute some evaluation statistics of the distance function
%                              assuming that the labels are provided (optional). If zero does nothing. default
%                              is set to 1, assuming labels are present, else set to zero. compares
%                              the distance function computed by DistBoost to various naive distance
%                              functions like: Euclidean, Whitened, and RCA.
%  
%         Knn          =   If evaluationHypFlag is on this is a row vector that sets various
%                          K-nearest-neighbor values for which the classification error will be
%                          computed. (optional). default value set to Knn=[1];
%
%     The function returns the following:
%         outputMats   =   Struct containing initial and final distance matrices of the training data.
%           outputMats.initial - initial distance matrix (t==1)
%           outputMats.final   - final distance matrix (t==T, or earlier exit).
% 
%         Hyp          =   Struct containing the Distance Function (or hypothesis) produced by DistBoost
%           Hyp.MogParams     -  The weak learner parameters (GMM parameters).
%           Hyp.hypWeight     -  The weights (or alphas in AdaBoost notation) of the different hyp.
%           Hyp.MarginType    -  The Margin Type used. typically set to 0 which means MAP margin (see initBSParams.m)
%
%         NeighborPurity =  If BSParams.evaluationHypFlag is set to 1 returns Cumulative Neighbor Purity
%                           statistics for the data (this also requires that the train labels be provided for
%                           evaluation purposes). also provides comparisons to three naive
%                           distance functions: Euclidean, Whitened, RCA and the initial distance
%                           matrix (T==1). this is therfore a
%                           4*K matrix in which each row is the cumulative neighbor purity of
%                           each of the methods (Euclidean, Whitened, RCA, DistBoost-initial hyp
%                           (t=1) and DistBoost-final hyp (t=T) respectively) where maxK is the
%                           maximal correct number of neighbors considered.  
%         
%         KnnError       =  If evaluationHypFlag is set to 1 and a Knn vector is passed computes
%                           the K-nearest Neighbor classifier error on the training data (this
%                           also requires that the train labels be provided for evaluation
%                           purposes). Also provides comparisons to three naive
%                           distance functions: Euclidean, Whitened, and RCA
%                           matrix. 
%                           This is therfore a 5*length(Knn) matrix in which each row is the Knn error for the
%                           values of Knn for each of the methods (Euclidean, Whitened, RCA,
%                           DistBoost-initial hyp (t=1), and DistBoost-final hyp (t=T)
%                           respectively) where maxK is the maximal correct number of neighbors
%                           considered.  
%
%         DevelopementStats = returns some boosting statistics (cumulative hyp loss, hypWeights, etc.)
% 
%----------------------------------------------------------------------------------------------------------

if(~exist('labels','var'))
  labels=[];
end

if(~exist('BSParams','var'))
  BSParams=[];
  BSParams.WeakParams=[];
end

%sets all uninitialized parameters to their default values, if such exist.
[BSParams]= initBSParams(BSParams);

[BSParams.WeakParams]= initWeakParams(BSParams.WeakParams,labels);

[N d]= size(data);
EMZs=zeros(3,BSParams.T);

[ PosInds , NegInds ]= createConstraintMatrix(chunklets,anti);

%ignore this!
global UnlabaledInds;   % used when simpleAlpha=2 in calcZ1

UnlabeledInds = setdiff(1:N^2,[ PosInds ; NegInds]);
ConstraintsNum=length([PosInds' NegInds']);
ConstraintsNumNoDiag=ConstraintsNum-N;

%initialize weights on pairs and points:
%allow passing weights from the outside!!
if(isfield(BSParams,'InitialPairW'))
  pairW= sparse(BSParams.InitialPairW);
  unlabeledW= BSParams.InitialUnlabeledW;
  pointW{1}= translateWeights(BSParams.InitialPairW,unlabeledW);
else
  pairW= sparse(N,N); % use sparse matrix to save up space/time.
  pairW([PosInds; NegInds])=1/(N*N);

  unlabeledW=1/(N*N);
  pointW{1}= ones(1,N);
  
end
finalHyp= zeros(N);%[]; %changed

repeat_counter=0; %how many attempts to produce a new hyp have been rejected due to low scores.
currNumberOfCalls=1; %how many attempts to produce new hyp have been tried in case of failure.
BadHypFlag=0; %if set, means last weak Hyp was badly correlated, and should be rerun.
DevelopementStats=[];

%main Boosting loop:
for t=1:BSParams.T    
  
  goodFlag=0;

  %loop on number of attempts to produce a new weak learner in each boosting round.
  for i=1:BSParams.numberOfCalls
    
    BSParams.WeakParams.w=pointW{t};
    [MogParams{t},ll ,assignment ,EMZs(:,t), acpChunks, exit_flags ]=...
	mixture_of_gaussians5(data, BSParams.WeakParams.M, chunklets, anti , BSParams.WeakParams);
    if(BSParams.printStatsFlag)
      sprintf('finished iteration %d out of %d iterations',t,BSParams.T)
    end
    
    if(isempty(MogParams{t})) %this means we failed to find an hyp.
      if(BSParams.printStatsFlag)
	disp(sprintf('iteration %d repeat %d returned empty GMM',t,i));
      end
      currNumberOfCalls= currNumberOfCalls +1;
	
    else %only if we got some hyp back..- now check its quality and update boosting parameters
	  
      [ probabilities, mapAssignment, marginMap]= calc_margin_map(data,MogParams{t},BSParams.MarginType);
      
      [ pairW,unlabeledW,hypWeight(t),finalHyp,hypCorr(t),stats{4}(t),stats{5}(t),BadHypFlag]= ...
	  boostConstraintsOpt(pairW,unlabeledW,marginMap,PosInds,NegInds,finalHyp,BSParams);
      
      %check to see if hyp is bad, if so retrain another one (done numberOfCalls times)
      if (BadHypFlag)
	if(BSParams.printStatsFlag)
	  disp(sprintf('iteration %d repeat %d failed',t,i));
	end
	repeat_counter=repeat_counter+1;
      else %success
	goodFlag=1;
	break;
      end
	  
    end
  %end
  end % for loop on numberOfCalls
    
  if(~goodFlag) %finish up boosting! case all attempts to get another weak learner failed:

    outputMats.final=finalHyp;
    %handle the case in which we failed on first iteration!!!!
    if(t==1)
      outputMats.initial= finalHyp;
    end
    stats{7}=[ 0,0 ;0 0 ; 0 0];
    assignment=[];
    if(BSParams.printStatsFlag)
      disp('Weak learner failed to find ahypothesis. Exit boosting');
    end
    break;
  else %iteration was successful
    [pointW{t+1}]= translateWeights(pairW,unlabeledW);
    
    if(t==1) %save initial hyp.
      outputMats.initial= finalHyp;
    end
    
    % gather current Hypothsis record. (will only hold the true finalHyp in the end!).
    outputMats.final= finalHyp;

    if(BSParams.evolutionFlag)
      outputMats.evolution{t}= finalHyp;
    end
    
    % gather additional boosting stats
    [DevelopementStats]= collectStats(DevelopementStats,t,finalHyp,PosInds,NegInds,hypWeight,hypCorr,EMZs,ConstraintsNum,ConstraintsNumNoDiag);
    
    %print supervised loss: sum of exp margins on sup. pairs.
    if(BSParams.printStatsFlag)
      disp(sprintf('iteration %d supervised loss %f',t,DevelopementStats{15}(t)));
    end
  
  end % of goodFlag==1


end % of Boosting loop

%save current Hyp  or distance function:
Hyp.MogParams= MogParams;
Hyp.hypWeight= DevelopementStats{1};
Hyp.MarginType= BSParams.MarginType;

%check whether to output similarity values or distance values (both bounded between 0 to 1).
if(BSParams.similarityFlag)
  outputMats= getSimilarityMats(outputMats);

else %distance matrices
  outputMats= getDistanceMats(outputMats);

end
  
%now if BSParams.printStats flag is on - print boosting development stats to screen:
if(BSParams.printStatsFlag)
  printDStats(DevelopementStats);
end


%now collect some evluation scores of the distance function learned.
if(BSParams.evaluateHypFlag)
  visual_flag=0; %whether to have evaluation plots pop out.
  HypRecord{1}= outputMats.initial;
  HypRecord{2}= outputMats.final;
  [ NeighborPurity , KnnError ] = evaluateHyp2(data,labels, chunklets,HypRecord, BSParams.Knn,visual_flag ); 
else
  NeighborPurity=[];
  KnnError=[];
end   

 