%modified to handle small/large probs by using LogProbs.
%modified to compute different types of margins.

% calculate signed margin map assignment of points using the given GMM model given
% by param. margin map is the joint probability of the map assignment of the pairs of points to
% the different GMM models, i.e if point x1 is assigned to model j1 and point x2 is assigned to
% model j2 the the joint is p(x1,x2| x1=j1, x2=j2). sign represents map if marginMap(i,j)> 0
% points i, and j are assigned to the same model in the MAP assignment, else they are not in the
% same model.

%params:
% data - the original data each data point is a row.
% param - the GMM model used.
% M - the number of models in the GMM
% TypeFlag - Type of margin : 0 (default) - Map margin
%                             1 - probabilistic margin
%                             2 - margin as difference between MAP and the second
%                             best.

% singleCovMatFlag - if 1 all GMM's have a singel covariance matrix.

%returns:

% probabilities : matrix of n*M where each row is the probability of a given point to belong to
%                each of the M  models.

% mapAssignment: vector of n dim. with the unconstrained MAP assignment of points to models (withouth any constraint
%                information used).
%               the mapAssignment argument is only valid when TypeFlag=0.
%               else [] is returned.

% marginMap:     matrix of n*n which contains the pairwise confidence of the MAP assignment. if
%                pos then points belong together, if neg, points do not belong together in MAP assignment.

function [ probabilities, mapAssignment,marginMap]= calc_margin_map(data,param,TypeFlag)

[M , stam ]= size(param);
singleCovMatFlag=1;
if isempty(param(2,2))
    singleCovMatFlag=0;
end
if ~exist('TypeFlag')
    TypeFlag=0;
end
[rows cols]= size(data);
n= rows;
d= cols;

%in order to cope with models that have been removed.
M= (size(param,1));

for j=1:M
    if (singleCovMatFlag == 1)
        cov_index=1;
    else
        cov_index=j;
    end
    
    tempData=data-ones(n,1)*param{j,1}; % subtract the mean
    q1=tempData*inv(param{cov_index,2}).*tempData;
    expContent=-0.5*sum(q1,2);         % the exponent content
    LogProbs(:,j)= -(d/2)*log(2*pi) -0.5*log(det(param{cov_index,2}))+expContent;
    LogProbs(:,j)=LogProbs(:,j)+log(param{j,3});
    
end

m=max(LogProbs');
LogProbs=LogProbs- m'*ones(1,M);
probabilities=exp(LogProbs);

%normalize probabilities:
probabilities= probabilities./(sum(probabilities')'*ones(1,M));

if TypeFlag==0
    [ tempVec, mapAssignment ]=max(probabilities');
    marginMap= tempVec'*tempVec; 
    for i=1:n   %take care of signs.
        negInds= find(mapAssignment ~= mapAssignment(i));
        marginMap(i,negInds)= -marginMap(i,negInds);
    end
elseif TypeFlag==1
    marginMap=probabilities*probabilities';
    marginMap=(marginMap-0.5)*2;
    mapAssignment=[];
elseif TypeFlag==2
    mult1 = repmat(probabilities',M,1);
    for i=1:n
        mult2=reshape(repmat(probabilities(i,:),M,1),M^2,1)*ones(1,n);
        pairProbs=mult1.*mult2;
        plusInds=1:M+1:M^2;
        MinusInds=setdiff(1:M^2,plusInds);
        BestPlus=max(pairProbs(plusInds,:));
        BestMinus=max(pairProbs(MinusInds,:));
        marginMap(i,:)=BestPlus-BestMinus;
    end
    mapAssignment=[];
else
    [ tempVec, mapAssignment ]=max(probabilities');
    marginMap=ones(n);
    for i=1:n   %take care of signs.
        negInds= find(mapAssignment ~= mapAssignment(i));
        marginMap(i,negInds)= -1;
    end
end
